import json
import logging
from pathlib import Path

import httpx

from yappa.handlers.wsgi import load_app, patch_response
from yappa.settings import DEFAULT_CONFIG_FILENAME
from yappa.utils import load_yaml

logger = logging.getLogger(__name__)


async def call_app(app, event):
    host_url = event["headers"].get("Host", "https://raw-function.net")
    if not host_url.startswith("http"):
        host_url = f"https://{host_url}"
    async with httpx.AsyncClient(app=app,
                                 base_url=host_url) as client:
        request = client.build_request(
            method=event["httpMethod"],
            url=event["url"],
            headers=event["headers"],
            params=event["queryStringParameters"],
            content=json.dumps(event["body"]).encode(),
        )
        response = await client.send(request)
        return response


try:
    config = load_yaml(Path(Path(__file__).resolve().parent.parent,
                            DEFAULT_CONFIG_FILENAME))

    app = load_app(config.get("entrypoint"),
                   config.get("DJANGO_SETTINGS_MODULE"))
except ValueError:
    logger.warning("Looks like broken Yappa config is used")


async def handle(event, context):
    if not event:
        return {
            'statusCode': 500,
            'body': "got empty event",
        }
    response = await call_app(app, event)
    return patch_response(response)



import json
import logging
from datetime import datetime

import click
from google.protobuf.empty_pb2 import Empty
from pytz import utc
from yandex.cloud.access.access_pb2 import (
    AccessBinding,
    AccessBindingDelta, Subject, UpdateAccessBindingsMetadata,
    UpdateAccessBindingsRequest,
)
from yandex.cloud.iam.v1.awscompatibility.access_key_service_pb2 import \
    CreateAccessKeyRequest
from yandex.cloud.iam.v1.awscompatibility.access_key_service_pb2_grpc import \
    AccessKeyServiceStub
from yandex.cloud.iam.v1.iam_token_service_pb2_grpc import IamTokenServiceStub
from yandex.cloud.iam.v1.key_service_pb2 import (
    CreateKeyRequest,
    DeleteKeyMetadata, DeleteKeyRequest,
)
from yandex.cloud.iam.v1.key_service_pb2_grpc import KeyServiceStub
from yandex.cloud.iam.v1.service_account_pb2 import ServiceAccount
from yandex.cloud.iam.v1.service_account_service_pb2 import \
    (
    CreateServiceAccountMetadata, CreateServiceAccountRequest,
    ListServiceAccountsRequest,
)
from yandex.cloud.iam.v1.service_account_service_pb2_grpc import \
    ServiceAccountServiceStub
from yandex.cloud.resourcemanager.v1.cloud_service_pb2 import ListCloudsRequest
from yandex.cloud.resourcemanager.v1.cloud_service_pb2_grpc import \
    CloudServiceStub
from yandex.cloud.resourcemanager.v1.folder_service_pb2 import \
    ListFoldersRequest
from yandex.cloud.resourcemanager.v1.folder_service_pb2_grpc import \
    FolderServiceStub
from yandexcloud._auth_fabric import get_auth_token_requester

from yappa.settings import DEFAULT_ACCESS_KEY_FILE, DEFAULT_SERVICE_ACCOUNT


def save_key(access_key, output_filename=DEFAULT_ACCESS_KEY_FILE):
    """
    saves access key to .yc json file
    """
    with open(output_filename, "w+") as f:
        f.write(json.dumps(access_key, indent=4))


logger = logging.getLogger(__name__)


class YcAccessMixin:
    sdk = None

    def _get_service_accounts(self):
        return self.sdk.client(ServiceAccountServiceStub).List(
            ListServiceAccountsRequest(
                folder_id=self.folder_id
            )).service_accounts

    def create_service_account(self,
                               service_account_name=DEFAULT_SERVICE_ACCOUNT,
                               ):
        """
        if there is not account with such name, creates such account
        """
        for account in self._get_service_accounts():
            if account.name == service_account_name:
                logger.warning("Account %s already exists, skipping creation",
                               service_account_name)
                return account
        click.echo("Creating service account "
                   + click.style(service_account_name, bold=True))
        account = self.sdk.wait_operation_and_get_result(
            self.sdk.client(ServiceAccountServiceStub).Create(
                CreateServiceAccountRequest(
                    folder_id=self.folder_id,
                    name=service_account_name,
                    description="Yappa service account: upload "
                                "to s3, "
                                "create functions and gateways",
                )),
            response_type=ServiceAccount,
            meta_type=CreateServiceAccountMetadata,
        ).response
        self.sdk.wait_operation_and_get_result(
            self.sdk.client(FolderServiceStub).UpdateAccessBindings(
                UpdateAccessBindingsRequest(
                    resource_id=self.folder_id,
                    access_binding_deltas=[
                        AccessBindingDelta(
                            action="ADD",
                            access_binding=AccessBinding(
                                role_id="editor",
                                subject=Subject(
                                    id=account.id,
                                    type="serviceAccount")),
                        ),
                        AccessBindingDelta(
                            action="ADD",
                            access_binding=AccessBinding(
                                role_id="serverless.functions.admin",
                                subject=Subject(
                                    id=account.id,
                                    type="serviceAccount")),
                        ),

                    ]
                )),
            response_type=Empty,
            meta_type=UpdateAccessBindingsMetadata,
        )
        return account

    def get_s3_key(self, service_account_name=DEFAULT_SERVICE_ACCOUNT,
                   description="yappa upload"):
        """
        if YC was instantiated with OAuth token: service account id is used
        else: it means we are already using service account, hence generate
        access key for using service account
        """
        if not self.service_account_id:
            self.service_account_id = self.create_service_account(
                service_account_name).id
        response = self.sdk.client(AccessKeyServiceStub).Create(
            CreateAccessKeyRequest(
                service_account_id=self.service_account_id,
                description=description))
        return {
            "aws_access_key_id": response.access_key.key_id,
            "aws_secret_access_key": response.secret
        }

    def create_service_account_key(self, service_account_id):
        response = self.sdk.client(KeyServiceStub).Create(
            CreateKeyRequest(
                service_account_id=service_account_id,
            )
        )
        return {
            "id": response.key.id,
            "service_account_id": response.key.service_account_id,
            "created_at": datetime.fromtimestamp(
                response.key.created_at.seconds
                + response.key.created_at.nanos * 10e-9,
                tz=utc).isoformat(),
            "key_algorithm": "RSA_2048",
            "public_key": response.key.public_key,
            "private_key": response.private_key
        }

    def delete_key(self, key_id):
        self.sdk.wait_operation_and_get_result(
            self.sdk.client(KeyServiceStub).Delete(
                DeleteKeyRequest(
                    key_id=key_id
                )),
            response_type=Empty,
            meta_type=DeleteKeyMetadata,
        )

    def get_clouds(self):
        return self.sdk.client(CloudServiceStub).List(
            ListCloudsRequest()).clouds

    def get_folders(self, cloud_id):
        return self.sdk.client(FolderServiceStub).List(
            ListFoldersRequest(cloud_id=cloud_id)).folders

    def get_iam_token(self):
        request = get_auth_token_requester(self.token,
                                           self.service_account_key).get_token_request()
        return self.sdk.client(IamTokenServiceStub).Create(request).iam_token



import logging
import os
import subprocess
import sys
from contextlib import suppress
from pathlib import Path
from shutil import copytree, ignore_patterns, make_archive, rmtree

import boto3

from yappa.settings import (
    DEFAULT_CONFIG_FILENAME, DEFAULT_IGNORED_FILES,
    DEFAULT_PACKAGE_DIR,
    DEFAULT_REQUIREMENTS_FILE,
    HANDLERS_DIR, YANDEX_S3_URL,
)

logger = logging.getLogger(__name__)


def prepare_package(requirements_file=DEFAULT_REQUIREMENTS_FILE,
                    ignored_files=DEFAULT_IGNORED_FILES,
                    tmp_dir=DEFAULT_PACKAGE_DIR, to_install_requirements=True,
                    config_filename=DEFAULT_CONFIG_FILENAME):
    """
    prepares package folder
    - copy project files
    - copy handler.py
    - install packages
    """
    logger.info('Copying project files to %s', tmp_dir)
    with suppress(FileExistsError):
        os.mkdir(tmp_dir)
    copytree(os.getcwd(), tmp_dir,
             ignore=ignore_patterns(*ignored_files, tmp_dir),
             dirs_exist_ok=True)
    copytree(Path(Path(__file__).resolve().parent, HANDLERS_DIR),
             Path(tmp_dir, "handlers"), dirs_exist_ok=True)
    os.rename(Path(tmp_dir, config_filename),
              Path(tmp_dir, DEFAULT_CONFIG_FILENAME))
    if to_install_requirements:
        logger.info('Installing requirements...')
        cmd = (
            f'{sys.executable} -m pip install '
            f'-r {requirements_file} -t {tmp_dir} --upgrade --quiet'
        )
        subprocess.check_call(cmd.split())
    return tmp_dir


def ensure_bucket(bucket_name, aws_access_key_id, aws_secret_access_key):
    s3 = boto3.resource(
        's3',
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        endpoint_url=YANDEX_S3_URL,
    )
    bucket = s3.Bucket(bucket_name)
    try:
        bucket.create()
    except Exception as e:
        if e.__class__.__name__ != 'BucketAlreadyOwnedByYou':
            raise
    return bucket


def upload_to_bucket(folder, bucket_name, aws_access_key_id,
                     aws_secret_access_key):
    """
    makes archive, uploads to bucket, deletes tmp archive
    """
    logger.info("Creating zip package")
    archive_path = make_archive(folder, 'zip', folder)
    archive_filename = os.path.basename(archive_path)
    try:
        bucket = ensure_bucket(bucket_name,
                               aws_access_key_id=aws_access_key_id,
                               aws_secret_access_key=aws_secret_access_key, )
        bucket.upload_file(archive_filename, archive_filename)
    finally:
        os.remove(archive_filename)
        rmtree(folder)
    return archive_filename


def delete_bucket(bucket_name, aws_access_key_id, aws_secret_access_key):
    """
    deletes bucket from s3
    """
    bucket = ensure_bucket(bucket_name, aws_access_key_id=aws_access_key_id,
                           aws_secret_access_key=aws_secret_access_key, )
    bucket.objects.all().delete()
    bucket.delete()



import yaml

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from yappa.settings import HANDLERS

MIN_MEMORY, MAX_MEMORY = 134217728, 2147483648

SIZE_SUFFIXES = {
    'kb': 1024,
    'mb': 1024 * 1024,
    'gb': 1024 * 1024 * 1024,
}


def convert_size_to_bytes(size_str):
    for suffix in SIZE_SUFFIXES:
        if size_str.lower().endswith(suffix):
            size = int(size_str[0:-len(suffix)]) * SIZE_SUFFIXES[suffix]
            if not MIN_MEMORY <= size <= MAX_MEMORY:
                raise ValueError("Sorry. Due to YandexCloud limits, function "
                                 "memory should be between 128mB and 2GB")
            return size
    raise ValueError("Oops. Couldn't parse memory limit. "
                     "It should be in format 128MB, 2GB")


def get_yc_entrypoint(application_type, raw_entrypoint):
    entrypoint = HANDLERS.get(application_type)
    if application_type == "raw":
        entrypoint = raw_entrypoint
    if not entrypoint:
        raise ValueError(
            f"Sorry, supported app types are: "
            f"{','.join(HANDLERS.keys())}. "
            f"Got {application_type}"
        )
    return entrypoint


def load_yaml(file, safe=False):
    try:
        with open(file, "r") as f:
            return yaml.load(f.read(), Loader)
    except FileNotFoundError:
        if safe:
            return dict()
        else:
            raise


def save_yaml(config, filename):
    with open(filename, "w+") as f:
        f.write(yaml.dump(config, sort_keys=False))
    return filename



import json
import logging
import os
from importlib import import_module
from pathlib import Path

import httpx

from yappa.settings import DEFAULT_CONFIG_FILENAME
from yappa.utils import load_yaml
from .handle_utils import (set_access_token, update_django_pg_connection)

logger = logging.getLogger(__name__)


def load_app(import_path, django_settings_module=None):
    if not import_path:
        raise ValueError("import_path should not be empty")
    os.environ["DJANGO_SETTINGS_MODULE"] = django_settings_module or ""
    *submodules, app_name = import_path.split(".")
    module = import_module(".".join(submodules))
    app = getattr(module, app_name)
    return app


def call_app(app, event):
    """
    call wsgi app
    see https://cloud.yandex.ru/docs/functions/concepts/function-invoke
    #response
    """
    host_url = event["headers"].get("Host", "https://raw-function.net")
    if not host_url.startswith("http"):
        host_url = f"https://{host_url}"
    with httpx.Client(app=app,
                      base_url=host_url) as client:
        request = client.build_request(
            method=event["httpMethod"],
            url=event["url"],
            headers=event["headers"],
            params=event["queryStringParameters"],
            json=json.loads(event["body"]) if event["body"] else None,
        )
        response = client.send(request)
        return response


try:
    set_access_token()
    config = load_yaml(Path(Path(__file__).resolve().parent.parent,
                            DEFAULT_CONFIG_FILENAME))

    app = load_app(config.get("entrypoint"),
                   config.get("django_settings_module"))
except ValueError:
    logger.warning("Couldn't load app. Looks like broken Yappa config is used")


def patch_response(response):
    """
    returns Http response in the format of
    {
     'status code': 200,
     'body': body,
     'headers': {}
    }
    """
    return {
        'statusCode': response.status_code,
        'headers': dict(response.headers),
        'body': response.content.decode(),
        'isBase64Encoded': False,
    }


def handle(event, context):
    set_access_token()
    if not event:
        raise ValueError("Got empty event")

    if config["django_settings_module"]:
        update_django_pg_connection(context.token["access_token"])
        # TODO check if it's necessary to update db settings
    response = call_app(app, event)
    return patch_response(response)



import logging
from contextlib import suppress
from typing import Iterable

from google.protobuf.duration_pb2 import Duration
from google.protobuf.empty_pb2 import Empty
from yandex.cloud.access.access_pb2 import (
    AccessBinding,
    ListAccessBindingsRequest,
    SetAccessBindingsMetadata, SetAccessBindingsRequest, Subject,
)
from yandex.cloud.serverless.functions.v1.function_pb2 import (
    Function,
    Package,
    Resources,
    Version,
)
from yandex.cloud.serverless.functions.v1.function_service_pb2 import (
    CreateFunctionMetadata, CreateFunctionRequest,
    CreateFunctionVersionMetadata, CreateFunctionVersionRequest,
    DeleteFunctionMetadata, DeleteFunctionRequest,
    GetFunctionVersionByTagRequest, ListFunctionsRequest,
)
from yandex.cloud.serverless.functions.v1.function_service_pb2_grpc import \
    FunctionServiceStub

from yappa.utils import convert_size_to_bytes

logger = logging.getLogger(__name__)


class YcFunctionsMixin:
    sdk = None
    folder_id = None

    def get_function(self, name) -> Function:
        """
        convenience method to get deployed function
        """

        for f in self._get_functions():
            if f.name == name:
                return f
        raise ValueError(f"Oops. Didn't find any function by name {name}")

    def _get_functions(self, filter_=None) -> Iterable[Function]:
        functions = self.sdk.client(FunctionServiceStub).List(
            ListFunctionsRequest(folder_id=self.folder_id,
                                 filter=filter_)).functions
        return functions

    def create_function(self, name, description="",
                        is_public=True) -> Function:
        """
        returns function and if it is new function
        """
        with suppress(ValueError):
            function = self.get_function(name)
            return function, False
        operation = self.sdk.client(FunctionServiceStub).Create(
            CreateFunctionRequest(
                folder_id=self.folder_id,
                name=name,
                description=description,
            ))
        operation_result = self.sdk.wait_operation_and_get_result(
            operation,
            response_type=Function,
            meta_type=CreateFunctionMetadata,
        )
        function = operation_result.response
        self.set_function_access(function.id, is_public)
        return function, True

    def delete_function(self, function_name):
        function = self.get_function(function_name)
        operation = self.sdk.client(FunctionServiceStub).Delete(
            DeleteFunctionRequest(
                function_id=function.id
            ))
        operation_result = self.sdk.wait_operation_and_get_result(
            operation,
            response_type=Function,
            meta_type=DeleteFunctionMetadata,
        )
        return operation_result.response

    def set_function_access(self, function_id=None, is_public=True,
                            function_name=None):
        """
        returns if access has changed
        """
        if function_name and not function_id:
            function_id = self.get_function(function_name).id
        if self._is_function_public(function_id) == is_public:
            return False
        if is_public:
            access_bindings = [AccessBinding(
                role_id='serverless.functions.invoker',
                subject=Subject(
                    id="allUsers",
                    type="system",
                )
            )]
        else:
            access_bindings = []
        operation = self.sdk.client(FunctionServiceStub).SetAccessBindings(
            SetAccessBindingsRequest(
                resource_id=function_id,
                access_bindings=access_bindings
            ))
        self.sdk.wait_operation_and_get_result(
            operation,
            response_type=Empty,
            meta_type=SetAccessBindingsMetadata,
        )
        return True

    def _is_function_public(self, function_id) -> bool:
        access_bindings = self.sdk.client(
            FunctionServiceStub).ListAccessBindings(
            ListAccessBindingsRequest(
                resource_id=function_id
            )).access_bindings
        for binding in access_bindings:
            if binding.role_id == 'serverless.functions.invoker':
                subject = binding.subject
                return subject.id == "allUsers"
        return False

    def create_function_version(self, function_name, runtime, description,
                                bucket_name, object_name,
                                entrypoint,
                                memory="128MB", service_account_id=None,
                                timeout=None, named_service_accounts=None,
                                environment=None) -> Version:
        function = self.get_function(function_name)
        operation = self.sdk.client(FunctionServiceStub).CreateVersion(
            CreateFunctionVersionRequest(
                function_id=function.id,
                runtime=runtime,
                description=description,
                entrypoint=entrypoint,
                resources=Resources(
                    memory=convert_size_to_bytes(memory)),
                execution_timeout=Duration(seconds=timeout),
                service_account_id=service_account_id,
                package=Package(bucket_name=bucket_name,
                                object_name=object_name),
                named_service_accounts=named_service_accounts,
                environment=environment
            ))
        operation_result = self.sdk.wait_operation_and_get_result(
            operation,
            response_type=Version,
            meta_type=CreateFunctionVersionMetadata,
        )
        return operation_result.response

    def get_latest_version(self, function_id) -> Version:
        version = self.sdk.client(FunctionServiceStub).GetVersionByTag(
            GetFunctionVersionByTagRequest(
                function_id=function_id,
                tag="$latest",
            ))
        return version



import logging
from contextlib import suppress
from typing import Iterable

import yandexcloud
from google.protobuf.empty_pb2 import Empty
from yandex.cloud.serverless.apigateway.v1.apigateway_pb2 import ApiGateway
from yandex.cloud.serverless.apigateway.v1.apigateway_service_pb2 import \
    (
    CreateApiGatewayMetadata, CreateApiGatewayRequest,
    DeleteApiGatewayMetadata,
    DeleteApiGatewayRequest,
    ListApiGatewayRequest,
)
from yandex.cloud.serverless.apigateway.v1.apigateway_service_pb2_grpc import \
    ApiGatewayServiceStub

yandexcloud._sdk._supported_modules[
    'yandex.cloud.serverless.apigateway'] = 'serverless-apigateway'
# TODO remove monkey patch when update is in PIP
logger = logging.getLogger(__name__)


class YcGatewayMixin:
    sdk = None
    folder_id = None

    def get_gateway(self, name) -> ApiGateway:
        """
         convenience method to get deployed gateway
         """
        for gw in self._get_gateways():
            if gw.name == name:
                return gw
        raise ValueError(f"Oops. Didn't find any gateway by name {name}")

    def _get_gateways(self, filter_=None) -> Iterable[ApiGateway]:
        gateways = self.sdk.client(ApiGatewayServiceStub).List(
            ListApiGatewayRequest(folder_id=self.folder_id,
                                  filter=filter_)).api_gateways
        return gateways

    def create_gateway(self, name, openapi_spec,
                       description="") -> ApiGateway:
        with suppress(ValueError):
            gateway = self.get_gateway(name)
            return gateway, False
        operation = self.sdk.client(ApiGatewayServiceStub).Create(
            CreateApiGatewayRequest(
                folder_id=self.folder_id,
                name=name,
                description=description,
                openapi_spec=openapi_spec
            ))
        operation_result = self.sdk.wait_operation_and_get_result(
            operation,
            response_type=ApiGateway,
            meta_type=CreateApiGatewayMetadata,
        )
        return operation_result.response, True

    def delete_gateway(self, gateway_name):
        gateway = self.get_gateway(gateway_name)
        operation = self.sdk.client(ApiGatewayServiceStub).Delete(
            DeleteApiGatewayRequest(
                api_gateway_id=gateway.id,
            ))
        self.sdk.wait_operation_and_get_result(
            operation,
            response_type=Empty,
            meta_type=DeleteApiGatewayMetadata,
        )

    def update_gateway(self, gateway_name, description,
                       openapi_spec) -> ApiGateway:
        logger.warning("Update gateway is not yet implemented")
        gateway = self.get_gateway(gateway_name)
        # TODO imlement update
        return gateway



import logging
import os

import httpx

logger = logging.getLogger(__name__)
TOKEN_URL = 'http://169.254.169.254/computeMetadata/v1/instance/service-accounts/default/token'


def set_access_token(iam_token=None):
    if not iam_token:
        iam_token = ""
        try:
            resp = httpx.get(TOKEN_URL, headers={'Metadata-Flavor': 'Google'})
            if resp.status_code == 200:
                iam_token = resp.json()["access_token"]
        except (httpx.ConnectError, httpx.ConnectTimeout) as e:
            logger.error("couldn't fetch IAM token: %s", e)
    os.environ["IAM_TOKEN"] = iam_token


def update_django_pg_connection(iam_token):
    from django.db import connections
    connections.databases["default"]["PASSWORD"] = iam_token
    # TODO not hard code db name



import io
import json
import logging
import os
from contextlib import redirect_stderr, redirect_stdout
from pathlib import Path

from yappa.settings import DEFAULT_CONFIG_FILENAME
from yappa.utils import load_yaml
from .handle_utils import set_access_token

logger = logging.getLogger(__name__)

try:
    config = load_yaml(
        Path(Path(__file__).resolve().parent.parent,
             DEFAULT_CONFIG_FILENAME))
    os.environ["DJANGO_SETTINGS_MODULE"] = config[
                                               "django_settings_module"] or ""
except KeyError:
    logger.error("DJANGO_SETTINGS_MODULE not present in the config")
except ValueError:
    logger.warning("Couldn't load app. Looks like broken Yappa config is used")

FORBIDDEN_COMMANDS = (
    'dbshell',
    'makemigrations',
    'runserver',
    'shell',
    'squashmigrations',
    'startapp',
    'startproject',
)

NOINPUT_COMMANDS = (
    "createsuperuser",
    "migrate",
)


def run_command(command, args):
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    from django.conf import settings
    logger.error(json.dumps(settings.DATABASES, indent=4))
    with io.StringIO() as buf, redirect_stdout(buf), redirect_stderr(buf):
        try:
            if not {"--noinput", "--no-input"}.intersection(args) \
                    and command in NOINPUT_COMMANDS:
                args = [*args, "--no-input"]
            execute_from_command_line(["__main__.py", command, *args])
            output = buf.getvalue()
        except Exception as e:
            output = str(e)
    return output


def manage(event, context=None):
    if context:
        set_access_token(context.token["access_token"])
    body = json.loads(event["body"])
    output_buffer = run_command(body["command"],
                                body.get("args") or [])
    return {
        'statusCode': 200,
        'body': output_buffer,
    }



import logging
from contextlib import suppress

import click
from click import ClickException

from yappa.cli_helpers import (
    NaturalOrderGroup, create_function_version, create_gateway,
    ensure_function, get_missing_details, update_gateway,
)
from yappa.config_generation import (
    create_default_config,
)
from yappa.handlers.manage import FORBIDDEN_COMMANDS
from yappa.s3 import delete_bucket
from yappa.settings import (
    DEFAULT_ACCESS_KEY_FILE, DEFAULT_CONFIG_FILENAME,
    YANDEX_OAUTH_URL,
)
from yappa.utils import load_yaml, save_yaml
from yappa.yc import YC
from yappa.yc.access import save_key
from yappa.yc.function_call import call_manage_function

logger = logging.getLogger(__name__)


@click.group(cls=NaturalOrderGroup)
def cli():
    pass


@cli.command(short_help="setup YC access")
@click.argument("config-file", type=click.Path(exists=False),
                default=DEFAULT_CONFIG_FILENAME, )
@click.option("-t", '--token', default="",
              help="yandex OAuth token", envvar="YC_OAUTH"
              )
def setup(config_file, token):
    """
    setup of cloud access:

    \b
      - asks for OAuth token
      - asks for cloud to work with
      - asks for folder to work with
      - creates service account (editor) and saves access key to .yc
      - saves folder_id and s3_account_name to yappa.yaml


    *you can skip this step if YC_OAUTH and YC_FOLDER in env vars
    (see README for authentication details)
    """
    click.echo("Welcome to " + click.style("Yappa", fg="yellow") + "!")
    if not token:
        click.echo(f"Please obtain OAuth token at "
                   + click.style(YANDEX_OAUTH_URL, fg="yellow"))
        token = click.prompt("Please enter OAuth token")
    yc = YC.setup(token=token, skip_folder=True)
    clouds = {c.name: c.id for c in yc.get_clouds()}
    cloud_name = click.prompt("Please select cloud", type=click.Choice(clouds),
                              default=next(reversed(clouds)))
    folders = {f.name: f.id for f in yc.get_folders(clouds[cloud_name])}
    folder_name = click.prompt("Please select folder",
                               type=click.Choice(folders),
                               default=next(reversed(folders)))
    yc.folder_id = folders[folder_name]
    account = yc.create_service_account(f"yappa-creator-account-{folder_name}")
    save_key(yc.create_service_account_key(account.id))
    click.echo("Saved service account credentials at " + click.style(
        DEFAULT_ACCESS_KEY_FILE, bold=True))

    config = (load_yaml(config_file, safe=True)
              or create_default_config(config_file))
    config["folder_id"] = folders[folder_name]
    config["service_account_names"]["creator"] = account.name
    save_yaml(config, config_file)
    click.echo("Saved Yappa config file at "
               + click.style(config_file, bold=True))


@cli.command(short_help='generate config files, create function & api-gateway')
@click.argument("config-file", type=click.Path(exists=False),
                default=DEFAULT_CONFIG_FILENAME, )
def deploy(config_file):
    """\b
    - generates yappa.yaml (skipped if file exists)
    - creates function
    - creates function version (with uploaded to s3 package)
    - generates yappa-gw.yaml (skipped if file exists)
    - creates api-gateway
    """
    config = (load_yaml(config_file, safe=True)
              or create_default_config(config_file))
    config, is_updated = get_missing_details(config)
    if is_updated:
        save_yaml(config, config_file)
        click.echo("saved Yappa config file at "
                   + click.style(config_file, bold=True))
    yc = YC.setup(config=config)
    function = ensure_function(yc, config["project_slug"],
                               config["description"], True)
    if config['django_settings_module']:
        ensure_function(yc, config["manage_function_name"],
                        config["description"], False)
    create_function_version(yc, config)
    if config["gw_config"]:
        is_new = create_gateway(yc, config, function.id)
        if not is_new:
            update_gateway(yc, config)


@cli.command()
@click.argument("config-file", type=click.Path(exists=True),
                default=DEFAULT_CONFIG_FILENAME, )
def undeploy(config_file):
    """
    deletes function, api-gateway and bucket
    """
    config = load_yaml(config_file)
    yc = YC.setup(config=config)
    with suppress(ValueError):
        click.echo(f"Deleting function {config['project_slug']}...")
        yc.delete_function(config["project_slug"])

    with suppress(ValueError):
        click.echo(f"Deleting api-gateway {config['project_slug']}...")
        yc.delete_gateway(config['project_slug'])
    with suppress(ValueError):
        click.echo(f"Destroying bucket {config['bucket']}...")
        delete_bucket(config["bucket"], **yc.get_s3_key(
            config["service_account_names"]["creator"]))
    click.echo("That's it! Only service account is left.\n"
               + click.style("Bye!", fg="yellow"))


@cli.command(context_settings=dict(ignore_unknown_options=True,
                                   help_option_names=[], ))
@click.option("--config-file", type=click.Path(exists=True),
              default=DEFAULT_CONFIG_FILENAME, )
@click.argument("command", type=str)
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def manage(config_file, command, args):
    if command in FORBIDDEN_COMMANDS:
        raise ClickException("Sorry. You cannot run this command with "
                             "Serverless setup")
    config = load_yaml(config_file, safe=False)
    yc = YC.setup(config=config)
    function = yc.get_function(config["manage_function_name"])
    response = call_manage_function(yc, function.id, command, args)
    click.echo(response)


if __name__ == '__main__':
    cli()



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    README = f.read()

setup(
    name='yappa',
    version='0.4.14',
    url='https://github.com/turokg/yappa',
    description='Easy serverless deploy of python web applications',
    long_description_content_type="text/markdown",
    long_description=README,
    author='Egor Korovin',
    author_email='kbkor@yandex.ru',
    packages=find_packages(),
    install_requires=[
        'boto3>=1.10',
        'click>=8.0',
        'httpx>=0.18',
        'yandexcloud>=0.92',
        'boltons>=21.0',
        'idna<3,>=2.5',
        "PyYAML>=5.0",
        "furl>=2.0",
        "pytz>=2021"
    ],
    python_requires='>=3.8.0',
    entry_points={'console_scripts': ['yappa = yappa.cli:cli']},
    license="MIT",
    package_data={'yappa': ['*.yaml']},
    include_package_data=True,
)



from uuid import uuid4

import click
import yaml
from boltons.strutils import slugify
from click import ClickException

from yappa.config_generation import (
    create_default_gw_config,
    inject_function_id,
)
from yappa.s3 import prepare_package, upload_to_bucket
from yappa.settings import HANDLERS
from yappa.utils import get_yc_entrypoint, load_yaml, save_yaml


class NaturalOrderGroup(click.Group):

    def list_commands(self, ctx):
        return self.commands.keys()


def ensure_function(yc, name, description, is_public):
    click.echo("Ensuring function...")
    function, is_new = yc.create_function(name,
                                          description, is_public=is_public)
    if is_new:
        click.echo("Created serverless function:\n\tname: "
                   + click.style(f"{function.name}")
                   + "\n\tid: "
                   + click.style(f"{function.id}")
                   + "\n\tinvoke url : "
                   + click.style(f"{function.http_invoke_url}", fg="yellow"))
    else:
        click.echo("Using existing function:\n\tname: "
                   + click.style(f"{function.name}")
                   + "\n\tid: "
                   + click.style(f"{function.id}")
                   + "\n\tinvoke url : "
                   + click.style(f"{function.http_invoke_url}", fg="yellow"))
    return function


def create_function_version(yc, config):
    click.echo("Preparing package...")
    package_dir = prepare_package(config["requirements_file"],
                                  config["excluded_paths"],
                                  to_install_requirements=True,
                                  )
    click.echo(f"Uploading to bucket {config['bucket']}...")
    object_key = upload_to_bucket(package_dir, config["bucket"],
                                  **yc.get_s3_key(
                                      config["service_account_names"][
                                          "creator"]))
    click.echo(f"Creating new function version for "
               + click.style(config["project_slug"], bold=True))
    yc.create_function_version(
        config["project_slug"],
        runtime=config["runtime"],
        description=config["description"],
        bucket_name=config["bucket"],
        object_name=object_key,
        entrypoint=get_yc_entrypoint(config["application_type"],
                                     config["entrypoint"]),
        memory=config["memory_limit"],
        service_account_id=config["service_account_id"],
        timeout=config["timeout"],
        named_service_accounts=config["named_service_accounts"],
        environment=config["environment"],
    )
    click.echo(f"Created function version")
    access_changed = yc.set_function_access(
        function_name=config["project_slug"], is_public=config["is_public"])
    if access_changed:
        click.echo(f"Changed function access. Now it is "
                   f" {'not' if config['is_public'] else 'open to'} public")

    if config["django_settings_module"]:
        yc.create_function_version(
            config["manage_function_name"],
            runtime=config["runtime"],
            description=config["description"],
            bucket_name=config["bucket"],
            object_name=object_key,
            entrypoint=get_yc_entrypoint("manage",
                                         config["entrypoint"]),
            memory=config["memory_limit"],
            service_account_id=config["service_account_id"],
            timeout=300,
            named_service_accounts=config["named_service_accounts"],
            environment=config["environment"],
        )


def create_gateway(yc, config, function_id):
    gw_config_filename = config["gw_config"]
    gw_config = (load_yaml(gw_config_filename, safe=True)
                 or create_default_gw_config(gw_config_filename))
    gw_config = inject_function_id(gw_config, f"{function_id}", config[
        "project_slug"])
    save_yaml(gw_config, gw_config_filename)
    click.echo("Saved Yappa Gateway config file at "
               + click.style(gw_config_filename, bold=True))
    click.echo("Ensuring api-gateway...")
    gateway, is_new = yc.create_gateway(config["project_slug"],
                                        yaml.dump(gw_config))
    if is_new:
        click.echo("Created api-gateway:\n\tname: "
                   + click.style(f"{gateway.name}")
                   + "\n\tid: "
                   + click.style(f"{gateway.id}", )
                   + "\n\tdomain : "
                   + click.style(f"https://{gateway.domain}", fg="yellow"))
    return is_new


def update_gateway(yc, config):
    gateway = yc.get_gateway(config["project_slug"])
    click.echo(f"Updating api-gateway "
               + click.style(f"{gateway.name}", bold=True))
    yc.update_gateway(gateway.name, config["description"],
                      load_yaml(config["gw_config"]))
    click.echo("Updated api-gateway:\n\tname: "
               + click.style(f"{gateway.name}")
               + "\n\tid: "
               + click.style(f"{gateway.id}", )
               + "\n\tdomain : "
               + click.style(f"https://{gateway.domain}", fg="yellow"))


class ValidationError(ClickException):
    pass


def is_valid_bucket_name(bucket_name):
    """
    Checks if an S3 bucket name is valid according to
    https://docs.aws.amazon.com/AmazonS3/latest/dev/BucketRestrictions.html
    """
    if len(bucket_name) < 3 or len(bucket_name) > 63:
        raise ValidationError("Bucket names must be at least 3 and no more "
                              "than 63 characters long.")
    if bucket_name.lower() != bucket_name or "_" in bucket_name:
        raise ValidationError("Bucket names must not contain uppercase"
                              " characters or underscores")
    for label in bucket_name.split("."):
        if len(label) < 1 \
                or not (label[0].islower() or label[0].isdigit()) \
                or not (label[-1].islower() or label[-1].isdigit()):
            raise ValidationError("Each label must start and end with a "
                                  "lowercase letter or a number")
    if all([s.isdigit() for s in bucket_name.split(".")]):
        raise ValidationError("Bucket names must not be formatted as an "
                              "IP address (i.e. 192.168.5.4)")


def is_valid_entrypoint(entrypoint):
    """
    try to import entrypoint. if is callable, then ok
    """


def is_valid_django_settings_module(django_settings_module):
    """
    try to setup django app
    """


def is_valid_requirements_file(requirements_file):
    """
    try to open requirements. if it matches to re
    """


def get_bucket_name(config):
    """
    generates bucket name, i.e. Yappa Project -> yappa.bucket-32139
    """
    return config['project_slug'].replace("_", ".") + f"-{str(uuid4())[:8]}"


def is_not_empty(string):
    if not string or not string.strip():
        raise ValidationError("should not be empty")


def is_valid_slug(string):
    """
    is has "_" or spaces - raise ViolationError
    """


def get_slug(config):
    return slugify(config["project_name"]).replace("_", "-")


PROMPTS = (
    ("project_name", "My project", [is_not_empty],
     "What's your project name?"),
    ("project_slug", get_slug, [is_valid_slug],
     "What's your project slug?"),
    ("bucket", get_bucket_name, [is_not_empty, is_valid_bucket_name],
     "Please specify bucket name"),
    ("requirements_file", "requirements.txt", [is_not_empty,
                                               is_valid_requirements_file],
     "Please specify requirements file")
)


def get_missing_details(config):
    """
    if value is missing in config prompt user
    """
    is_updated = False
    for key, default, validators, question in PROMPTS:
        if config.get(key) is not None:
            continue
        is_updated = True
        default = default(config) if callable(default) else default
        value = click.prompt(question, default=default)
        for validator in validators:
            validator(value)
        config[key] = value
    if not config.get("application_type"):
        application_types = list(HANDLERS)
        application_types.remove("manage")
        config["application_type"] = click.prompt(
            "Please specify application type",
            default=application_types[0],
            type=click.Choice(application_types), )
    if not config.get("entrypoint"):
        config["entrypoint"] = click.prompt(
            "Please specify import path for application",
            default="wsgi.app")
    if not config.get("django_settings_module") \
            and config["application_type"] == "Django":
        config["django_settings_module"] = click.prompt(
            "Please specify your DJANGO_SETTINGS_MODULE",
            default="project.project.settings")
    if config["django_settings_module"]:
        config["manage_function_name"] = f"{config['project_slug']}-manage"
    return config, is_updated
